/*
 *  linux/include/linux/nfs_fs.h
 *
 *  Copyright (C) 1992  Rick Sladkey
 *
 *  OS-specific nfs filesystem definitions and declarations
 */

#ifndef _LINUX_NFS_FS_H
#define _LINUX_NFS_FS_H

#ifdef __KERNEL__
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/in.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <linux/sunrpc/sched.h>
#endif /* __KERNEL__ */

#include <linux/nfs.h>

/*
 * Enable debugging support for nfs client.
 * Requires RPC_DEBUG.
 */
#ifdef RPC_DEBUG
# define NFS_DEBUG
#endif

#define NFS_MAX_FILE_IO_BUFFER_SIZE	32768
#define NFS_DEF_FILE_IO_BUFFER_SIZE	4096

/*
 * The upper limit on timeouts for the exponential backoff algorithm.
 */
#define NFS_MAX_RPC_TIMEOUT		(6*HZ)
#define NFS_WRITEBACK_DELAY		(5*HZ)
#define NFS_WRITEBACK_LOCKDELAY		(60*HZ)
#define NFS_COMMIT_DELAY		(5*HZ)

/*
 * superblock magic number for NFS
 */
#define NFS_SUPER_MAGIC			0x6969
#define NFS_DENTRY_MAGIC		0x2439
#define NFS_FILE_MAGIC			0xA4F0

#ifdef __KERNEL__
/*
 * Convenience macros
 */
#define NFS_FH(dentry)			(&NFS_DENTRY(dentry)->fh)
#define NFS_DSERVER(dentry)		(&(dentry)->d_sb->u.nfs_sb.s_server)
#define NFS_SERVER(inode)		(&(inode)->i_sb->u.nfs_sb.s_server)
#define NFS_CLIENT(inode)		(NFS_SERVER(inode)->client)
#define NFS_PROTO(inode)		(NFS_SERVER(inode)->rpc_ops)
#define NFS_REQUESTLIST(inode)		(NFS_SERVER(inode)->rw_requests)
#define NFS_ADDR(inode)			(RPC_PEERADDR(NFS_CLIENT(inode)))
#define NFS_CONGESTED(inode)		(RPC_CONGESTED(NFS_CLIENT(inode)))
#define NFS_COOKIEVERF(inode)		((inode)->u.nfs_i.cookieverf)
#define NFS_READTIME(inode)		((inode)->u.nfs_i.read_cache_jiffies)
#define NFS_CACHE_CTIME(inode)		((inode)->u.nfs_i.read_cache_ctime)
#define NFS_CACHE_MTIME(inode)		((inode)->u.nfs_i.read_cache_mtime)
#define NFS_CACHE_ATIME(inode)		((inode)->u.nfs_i.read_cache_atime)
#define NFS_CACHE_ISIZE(inode)		((inode)->u.nfs_i.read_cache_isize)
#define NFS_NEXTSCAN(inode)		((inode)->u.nfs_i.nextscan)
#define NFS_CACHEINV(inode) \
do { \
	NFS_READTIME(inode) = jiffies - NFS_MAXATTRTIMEO(inode) - 1; \
} while (0)
#define NFS_ATTRTIMEO(inode)		((inode)->u.nfs_i.attrtimeo)
#define NFS_MINATTRTIMEO(inode) \
	(S_ISDIR(inode->i_mode)? NFS_SERVER(inode)->acdirmin \
			       : NFS_SERVER(inode)->acregmin)
#define NFS_MAXATTRTIMEO(inode) \
	(S_ISDIR(inode->i_mode)? NFS_SERVER(inode)->acdirmax \
			       : NFS_SERVER(inode)->acregmax)
#define NFS_ATTRTIMEO_UPDATE(inode)	((inode)->u.nfs_i.attrtimeo_timestamp)

#define NFS_FLAGS(inode)		((inode)->u.nfs_i.flags)
#define NFS_REVALIDATING(inode)		(NFS_FLAGS(inode) & NFS_INO_REVALIDATING)

#define NFS_FILEID(inode)	((inode)->u.nfs_i.fileid)
#define NFS_FSID(inode)		((inode)->u.nfs_i.fsid)

#define NFS_FILE(file)		((struct nfs_file *)(file)->private_data)
#define NFS_DENTRY(dentry)	((struct nfs_dentry*)(dentry)->d_fsdata)


/* Inode Flags */
#define NFS_USE_READDIRPLUS(inode)	((NFS_FLAGS(inode) & NFS_INO_ADVISE_RDPLUS) ? 1 : 0)
#define NFS_MONOTONE_COOKIES(inode)	((NFS_SERVER(inode)->flags & NFS_NONMONOTONE_COOKIES) ? 0 : 1)

/*
 * These are the default flags for swap requests
 */
#define NFS_RPC_SWAPFLAGS		(RPC_TASK_SWAPPER|RPC_TASK_ROOTCREDS)

/* Flags in the RPC client structure */
#define NFS_CLNTF_BUFSIZE	0x0001	/* readdir buffer in longwords */

#define NFS_RW_SYNC		0x0001	/* O_SYNC handling */
#define NFS_RW_SWAP		0x0002	/* This is a swap request */

/*
 * When flushing a cluster of dirty pages, there can be different
 * strategies:
 */
#define FLUSH_AGING		0	/* only flush old buffers */
#define FLUSH_SYNC		1	/* file being synced, or contention */
#define FLUSH_WAIT		2	/* wait for completion */
#define FLUSH_STABLE		4	/* commit to stable storage */


/*
 * Structure for dentry->d_fsdata;
 */
struct nfs_dentry {
	unsigned long		magic;		/* Magic number */
	struct nfs_fh		fh;		/* File handle */
	struct rpc_cred*	cred;		/* RPC Credentials */
};

static inline int
nfs_check_dentry(struct dentry *dentry)
{
	if (NFS_DENTRY(dentry)->magic != NFS_DENTRY_MAGIC) {
		printk(KERN_ERR "NFS: corrupt dentry structure!\n");
		return 0;
	}
	return 1;
}

/*
 * Structure for file->private_data;
 */
struct nfs_file {
	unsigned long		magic;		/* Magic number */
	struct rpc_cred*	cred;		/* RPC Credentials */
};

static inline int
nfs_check_file(struct file *file)
{
	if (NFS_FILE(file)->magic != NFS_FILE_MAGIC) {
		printk(KERN_ERR "NFS: corrupt file structure!\n");
		return 0;
	}
	return 1;
}

static inline
unsigned long page_offset(struct page *page)
{
        return page->offset;
}

static inline
unsigned long page_index(struct page *page)
{
	return page->offset >> PAGE_CACHE_SHIFT;
}

/*
 * Argument struct for decode_entry function
 */
struct nfs_entry {
	struct page *		page;
	__u64			ino;
	__u64			cookie,
				prev_cookie;
	const char *		name;
	unsigned int		len;
	int			eof;
	struct nfs_fh		fh;
	struct nfs_fattr	fattr;
	unsigned long		offset,
				prev;
};

/*
 * RPC procedure vector for NFSv2/NFSv3 demuxing
 */
struct nfs_rpc_ops {
	int	version;		/* Protocol version */

	int	(*getroot)(struct nfs_server *,
			struct nfs_fh *, struct nfs_fattr *);
	int	(*getattr)(struct dentry *, struct nfs_fattr *);
	int	(*setattr)(struct dentry *, struct nfs_fattr *, struct iattr *);
	int	(*lookup)(struct dentry *, struct nfs_fattr *, struct qstr *,
			struct nfs_fh *, struct nfs_fattr *);
	int	(*access)(struct dentry *, int fmode, struct nfs_fattr *, int);
	int	(*readlink)(struct dentry *, struct nfs_fattr *,
			void *buffer, unsigned int buflen);
	int	(*read)(struct dentry *, struct nfs_fattr *,
			struct rpc_cred *,
			int flags, unsigned long offset,
			unsigned int count, void *buffer, int *eofp);
	int	(*write)(struct dentry *, struct nfs_fattr *,
			struct rpc_cred *,
			int flags, unsigned long offset,
			unsigned int count, void *buffer,
			struct nfs_writeverf *verfp);
	int	(*commit)(struct dentry *, struct nfs_fattr *,
			struct rpc_cred *,
			unsigned long, unsigned int);
	int	(*create)(struct dentry *, struct nfs_fattr *,
			struct qstr *, struct iattr *, int flags,
			struct nfs_fh *, struct nfs_fattr *);
	int	(*remove)(struct dentry *, struct nfs_fattr *,
			  struct qstr *, struct rpc_cred *);
	int	(*rename)(struct dentry *, struct nfs_fattr *, struct qstr *,
			struct dentry *, struct nfs_fattr *, struct qstr *);
	int	(*link)(struct dentry *, struct nfs_fattr *,
			struct dentry *, struct nfs_fattr *, struct qstr *);
	int	(*symlink)(struct dentry *, struct nfs_fattr *, struct qstr *,
			struct qstr *, struct iattr *,
			struct nfs_fh *, struct nfs_fattr *);
	int	(*mkdir)(struct dentry *, struct nfs_fattr *, struct qstr *,
			struct iattr *, struct nfs_fh *, struct nfs_fattr *);
	int	(*rmdir)(struct dentry *, struct nfs_fattr *, struct qstr *);
	int	(*readdir)(struct dentry *, struct nfs_fattr *,
			struct rpc_cred *,
			u64 cookie, void *, unsigned int size, int plus);
	int	(*mknod)(struct dentry *, struct nfs_fattr *, struct qstr *,
			struct iattr *, dev_t,
			struct nfs_fh *, struct nfs_fattr *);
	int	(*statfs)(struct nfs_server *, struct nfs_fh *,
			struct nfs_fsinfo *);
	u32 *	(*decode_dirent)(u32 *, struct nfs_entry *, int plus);
};

/*
 * 	NFS_CALL(getattr, inode, (fattr));
 * into
 *	NFS_PROTO(inode)->getattr(fattr);
 */
#define NFS_CALL(op, inode, args)	NFS_PROTO(inode)->op args


/*
 * Function vectors etc. for the NFS client
 */
extern struct nfs_rpc_ops	nfs_v2_clientops;
extern struct nfs_rpc_ops	nfs_v3_clientops;
extern struct rpc_version	nfs_version2;
extern struct rpc_version	nfs_version3;
extern struct rpc_program	nfs_program;

/*
 * linux/fs/nfs/inode.c
 */
extern struct inode *nfs_fhget(struct dentry *, struct nfs_fh *,
			       struct nfs_fattr *);
extern struct super_block *nfs_read_super(struct super_block *, void *, int);
extern int	init_nfs_fs(void);
extern void	nfs_zap_caches(struct inode *);
extern int	nfs_revalidate(struct dentry *);
extern int	nfs_open(struct inode *, struct file *);
extern int	nfs_release(struct inode *, struct file *);
extern int	__nfs_revalidate_inode(struct dentry *);
extern int	nfs_refresh_inode(struct inode *, struct nfs_fattr *);
extern int	nfs_wait_on_inode(struct inode *, int flag);
extern void	nfs_unlock_inode(struct inode *);
extern int	nfs_update_atime(struct dentry *);


/*
 * linux/fs/nfs/file.c
 */
extern struct inode_operations nfs_file_inode_operations;
extern int	nfs_permission(struct inode *i, int msk);

/*
 * linux/fs/nfs/dir.c
 */
extern struct inode_operations nfs_dir_inode_operations;
extern struct dentry_operations nfs_dentry_operations;
extern struct nfs_dentry *nfs_fh_alloc(void);
extern void nfs_fh_free(struct nfs_dentry *);


/*
 * linux/fs/nfs/symlink.c
 */
extern struct inode_operations nfs_symlink_inode_operations;

/*
 * linux/fs/nfs/locks.c
 */
extern int nfs_lock(struct file *, int, struct file_lock *);

/*
 * linux/fs/nfs/write.c
 */
extern struct nfs_page*	nfs_find_request(struct inode *, struct page *);
extern void nfs_release_request(struct nfs_page *req);
extern int  nfs_writepage(struct file *, struct page *);
extern int  nfs_updatepage(struct file *, struct page *, const char *,
                        unsigned long, unsigned int, int);
/*
 * Try to write back everything synchronously (but check the
 * return value!)
 */
extern int  nfs_sync_file(struct inode *, struct file *, unsigned long, unsigned int, int);
extern int  nfs_flush_file(struct inode *, struct file *, unsigned long, unsigned int, int);
extern int  nfs_commit_file(struct inode *, struct file *, unsigned long, unsigned int, int);
extern int  nfs_flush_timeout(struct inode *, int);
extern int  nfs_commit_timeout(struct inode *, int);

static inline int
nfs_have_writebacks(struct inode *inode)
{
	return !list_empty(&inode->u.nfs_i.writeback);
}

static inline int
nfs_wb_all(struct inode *inode)
{
	int error = nfs_sync_file(inode, 0, 0, 0, FLUSH_WAIT);
	return (error < 0) ? error : 0;
}

/*
 * Write back all requests on one page - we do this before reading it.
 */
static inline int
nfs_wb_page(struct inode *inode, struct page* page)
{
	int error = nfs_sync_file(inode, 0, page_index(page), 1, FLUSH_WAIT | FLUSH_STABLE);
	return (error < 0) ? error : 0;
}

/*
 * Write back all pending writes for one user.. 
 */
static inline int
nfs_wb_file(struct inode *inode, struct file *file)
{
	int error = nfs_sync_file(inode, file, 0, 0, FLUSH_WAIT);
	return (error < 0) ? error : 0;
}

/*
 * Lock the page of an asynchronous request
 */
static __inline__ int
nfs_lock_page(struct page *page)
{
	return !test_and_set_bit(PG_locked, &page->flags);
}

static __inline__ void
nfs_unlock_page(struct page *page)
{
        clear_bit(PG_locked, &page->flags);
        wake_up(&page->wait);
}

/*
 * linux/fs/nfs/read.c
 */
extern int  nfs_readpage(struct file *, struct page *);
extern struct page *  nfs_grab_cache_page(struct inode *, unsigned long);

/*
 * linux/fs/nfs2xdr.c
 */
extern u32 *nfs_decode_dirent(u32 *, struct nfs_entry *, int);

/*
 * linux/fs/nfs2xdr.c
 */
extern u32 *nfs3_decode_dirent(u32 *, struct nfs_entry *, int);

/*
 * linux/fs/mount_clnt.c
 * (Used only by nfsroot module)
 */
extern int  nfs_mount(struct sockaddr_in *, char *, struct nfs_fh *);
extern int  nfs3_mount(struct sockaddr_in *, char *, struct nfs_fh *);

/*
 * inline functions
 */
static inline int
nfs_revalidate_inode(struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	if (time_before(jiffies,NFS_READTIME(inode)+NFS_ATTRTIMEO(inode)))
		return 0;
	return __nfs_revalidate_inode(dentry);
}

static inline off_t
nfs_size_to_off_t(__u64 size)
{
	return (size > (__u64)LONG_MAX) ? (off_t)LONG_MAX : (off_t) size;
}

static inline ino_t
nfs_fileid_to_ino_t(u64 fileid)
{
	ino_t ino = (unsigned long) fileid;
	if (sizeof(unsigned long) < sizeof(u64))
		ino ^= fileid >> (sizeof(u64)-sizeof(unsigned long)) * 8;
	return ino;
}

static inline time_t
nfs_time_to_secs(__u64 time)
{
	return (time_t)(time >> 32);
}

static __inline__ struct rpc_cred *
nfs_file_cred(struct file *file)
{
	if (!NFS_FILE(file) || !nfs_check_file(file)) {
		printk("nfs_file_cred: invalid file!\n");
		return NULL;
	}
	return NFS_FILE(file)->cred;
}

static __inline__ struct rpc_cred *
nfs_dentry_cred(struct dentry *dentry)
{

	if (!NFS_DENTRY(dentry) || !nfs_check_dentry(dentry))
		return NULL;
	return NFS_DENTRY(dentry)->cred;
}


/* NFS root */

extern int nfs_root_mount(struct super_block *sb);

#define nfs_wait_event(clnt, wq, condition)				\
({									\
	int __retval = 0;						\
	if (clnt->cl_intr) {						\
		sigset_t oldmask;					\
		rpc_clnt_sigmask(clnt, &oldmask);			\
		__retval = wait_event_interruptible(wq, condition);	\
		rpc_clnt_sigunmask(clnt, &oldmask);			\
	} else								\
		wait_event(wq, condition);				\
	__retval;							\
})

#endif /* __KERNEL__ */

/*
 * NFS debug flags
 */
#define NFSDBG_VFS		0x0001
#define NFSDBG_DIRCACHE		0x0002
#define NFSDBG_LOOKUPCACHE	0x0004
#define NFSDBG_PAGECACHE	0x0008
#define NFSDBG_PROC		0x0010
#define NFSDBG_XDR		0x0020
#define NFSDBG_FILE		0x0040
#define NFSDBG_ROOT		0x0080
#define NFSDBG_PARANOID		0x0100
#define NFSDBG_ALL		0xFFFF

#ifdef __KERNEL__
# undef ifdebug
# ifdef NFS_DEBUG
#  define ifdebug(fac)		if (nfs_debug & NFSDBG_##fac)
# else
#  define ifdebug(fac)		if (0)
# endif
# ifdef NFS_PARANOIA
#  define nfsparanoid(args...)	dfprintk(PARANOID, ##args)
# else
#  define nfsparanoid(args...)	do { } while (0)
# endif
#endif /* __KERNEL */

#endif
